/* -*- C++ -*- */

/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 2002 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "SOAP" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

/*
 *
 * @authors  Lilantha Darshana    (lilantha@erunway.com)
 *           Satheesh Thurairajah (satheesh@erunway.com)
 *			 
 *
 */

#include "Platform.hpp"
#include "Call.hpp"
#include "SOAPMsgContext.hpp"
#include "AxisException.hpp"
#include <iostream>
#include "Param.hpp"


/**
 * Create a Call object and initialize Message Context
 * 
 */

Call::Call() : m_pClientCtx(0)
{

}


/**
 * Clean up the object for shutting down the Web service.
 *
 */

Call::~Call()
{
	Cleanup();
}


/**
 * Sets the operation name associated with this Call object. The validity of 
 * the operation is not checked at this point.
 *
 * @param	p_pchOpName Name of the operation.
 */
void Call::SetOperationName(const char *p_pchOpName)
{
    if (p_pchOpName)
        m_strOpName = p_pchOpName;
}


/**
 * Sets the operation name associated Namespace URI with this Call object. The 
 * validity of the operation namespace is not checked at this point. 
 *
 * @param	p_pchOpNsURI Namespace URI of the operation.
 */

void Call::SetMethodNamespaceURI(const char *p_pchOpNsURI)
{
    if (p_pchOpNsURI)
        m_strMethodNsURI = p_pchOpNsURI;
}


/**
 * Sets the URL of the target Web Service.
 *
 * @param	p_pchUrl of the target Web Service
 */

void Call::SetTargetEndpointAddress(const char* p_pchUrl)
{
	m_ServerURL = p_pchUrl;
}


/**
 * Map a type for XML serialization & deserialization
 *
 *
 * @param p_sTypName Name that mapping is required as xsi:type QName of the 
 *					 associated XML type.
 * @param p_pSer	 A SerializerFactory factory which can create serializer
 *					 /deserializer instances for this type.
 */

void Call::AddSerializerFactory(const std::string& p_sTypName, SerializerFactory * p_pSer)
{
	if(!m_pClientCtx)
		m_pClientCtx = new SOAPMsgContext();

	m_pClientCtx->RegisterTypeMapping(p_sTypName, p_pSer);
}


/**
 * Sets the encoding style to the URL passed in.
 *
 * @param Namespace URI of the encoding to use.
 */

void Call::SetEncodingStyle(const std::string& p_sNsURI)
{
	m_EncStyle = p_sNsURI;
}


/**
 * Invokes the operation associated with this Call object using the
 * Method namespace & method name provided as the arguments.
 *
 * @param	p_sNs	  Method namespace URI
 * @param	p_sMethod Method name (required)
 *
 * @return	Result of the Invoke of remote method. The object handler is return
 *			as a void pinter which client required to exlisitly cast in to the
 *			required data type.
 * 
 */

void * Call::Invoke(const std::string& p_sNs, const std::string& p_sMethod)
{
	// set method name & namespace of the method.
	m_strMethodNsURI = p_sNs;
	m_strOpName		 = p_sMethod;

	return Invoke();
}


/**
 * Invokes the operation associated with this Call object using the
 * method name in method namespace already provided, Create RPCElements
 * as body element using the vector of params (if any) for the method.
 * All the exceptions of such calls are handle here; do we required to 
 * pass such exceptions to the Client????
 *
 * @return	Result of the Invoke of remote method. The object handler is return
 *			as a void pinter which client required to explisitly cast in to the
 *			required data type. dynamic_cast<..> is strongly suggested. Client 
 *			require to clean up any memory associated with this returned obj.
 *
 * @todo    Manupulation of other return values;
 * 
 */

void * Call::Invoke()
{
	try
	{
		// Make sure the method name is provided
		if(m_strOpName.empty()) throw AxisException("Cannot Invoke unspecified method ....", 1);

		// Build up the RPC body from method name, namespace & set of arguments; as SOAPBody
		RPCElement  *pBody = new RPCElement(m_strMethodNsURI, m_strOpName, m_vArgs);

		// Create message contex to handle SOAP requests
		if(!m_pClientCtx)
			m_pClientCtx = new SOAPMsgContext();

		//Set remote Web service address
		m_pClientCtx->SetServiceURL(m_ServerURL);

		// Invoke the Web service with RPC body elements
		m_pClientCtx->Invoke(pBody);

		// Get result of call to remote service as RPC body Element
		RPCElement *ret = (RPCElement*)m_pClientCtx->GetResults();

		// return deserialized response message to the caller		
		// according the the service description of output form -- metainfo
		return ret->GetReturnValue(); 
									  
	}
	catch(AxisException& ex)
	{
		std::cerr << ex.GetErrorMsg() << std::endl;
		Cleanup();
	}
	catch(...)
	{
		// Handle any sort of exception that could cause on calling remote sercvice.
		std::cerr << "Exception in WebService ....." << std::endl;
		Cleanup();
	}
	
	// Exception has occured
	return NULL;
}


/**
 * Clean up on an exception
 *
 */

void Call::Cleanup()
{
	delete m_pClientCtx;
	m_pClientCtx = NULL;
	int size;
	if(size = m_vArgs.size())
	{
		for (int i=0; i < size ; i++)
			delete m_vArgs[i];
		m_vArgs.clear();
	}
}


