/* -*- C++ -*- */

/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 2002 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "SOAP" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

/*
 *
 * @authors		Lilantha Darshana	 (lilantha@erunway.com)
 *				Satheesh Thurairajah (satheesh@erunway.com)
 *			
 *
 */

#if !defined(_AXIS_CALL_HPP)
#define _AXIS_CALL_HPP


#include <vector>
#include <string>
#include "Param.hpp"
#include "XMLType.hpp"

class SerializerFactory;
class SOAPMsgContext;

/**
 * RPC Dynamic Invocation interface.  This class should be used to actually 
 * invoke the Web Service from Client side. 
 * The stratergy of calling this interface is as follows:
 * Set up & configure a object of this class with following informations:
 *		
 * \li	TARGER_ENDPOINT - Where the Web service can be located ( :-( not location transparent) 
 * \li	NAMESPACE       - Namespace URI of remote method 
 * \li	METHOD_NAME		- Name of the method to be Invoke from the end point
 * \li	PARAMETERES		- Parameters (if any) for calling remote method
 * \li	SERIALIZATION	- If user defined structured data types are used, provide
 *						  serialization factory for the purpose of XML serialization 
 *						  & deserialization of data.
 * 
 * @todo      Support for WSDL document (on the constructor to the Service object) 
 *	
 * @brief     RPC Dynamic Invocation interface
 *
 */

class Call  
{
public:

	/// Mode of the each parameter: This is not supported yet???, default is set as IN
	enum Mode {IN_PARAM, OUT_PARAM, IN_OUT_PARAM};

public:
	Call();
	virtual ~Call();
	
	/**
     * Adds the specified parameter to the list of parameters for the
     * operation associated with this Call object. This method should be called by
	 * the client to pass parameters on the method; call should be made on this
	 * according to the order of parameters that should be passed on the method.
	 *
	 * Store each param in a vector as name-value pair with the data type 
	 * Qualified name according to XML Schema Part 2 Datatype.
     *
	 * @brief	Adds specified parameter to the list of parameters 
     *			for the operation associated with this Call object.
	 *
     * @param   p_pchName    Name that will be used for the parameter in the XML 
	 *						 representation.
     * @param   p_Value	     Value of the param, rpresented genericly by typename
     * @param   p_Mode	     one of IN_PARAM, OUT_PARAM, IN_OUT_PARAM
	 *
	 */

	template<typename T>
	void AddParameter(const char* p_pchName, T& p_Value, Mode p_Mode)
	{
		Param *newParam = new Param;

   		newParam->m_pValue = (void*)&p_Value;
		newParam->m_pType  = ax::XMLType<T>::GetType();
		newParam->m_Name   = p_pchName;

 		m_vArgs.push_back(newParam);
	}

	/**
     * Adds the specified parameter to the list of parameters for the
     * operation associated with this Call object. This method should be called by
	 * the client to pass parameters on the method; call should be made on this
	 * according to the order of parameters that should be passed on the method.
	 *
	 * Store each param in a vector as name-value pair with the data type 
	 * Qualified name accord XML Schema Part 2 Datatype. Param name is generated
	 * like arg0, arg1 etc. accordingh to it place in the vectore.
     *
 	 * @brief	Adds specified parameter without a name to the list of parameters 
     *			for the operation associated with this Call object.
	 *
     * @param	p_Value	     Value of the param, rpresented genericly by typename
     * @param	p_Mode	     one of IN_PARAM, OUT_PARAM, IN_OUT_PARAM
	 *
	 */

	template<typename T>
	void AddParameter(T& value, Mode mode)
	{
		Param *newParam = new Param;

		char buf[10];
		sprintf(buf, "%d", m_vArgs.size());

   		newParam->m_pValue = (void*)&value;
		newParam->m_pType  = ax::XMLType<T>::GetType();
		newParam->m_pType  += buf; // create names like arg0, arg1 etc.

 		m_vArgs.push_back(newParam);
	}

	///Invokes the operation associated with this Call object using the method name in namespace 
	void * Invoke(const std::string& p_sNs, const std::string& p_sMethod);

	/// Invokes the operation associated with this Call object using info aready provided
	void * Invoke();

	/// Set romote address of the Web Service
	void SetTargetEndpointAddress(const char* p_pchUrl);

	/// set the method namespace URI
	void SetMethodNamespaceURI(const char *p_pchOpNsURI);

	/// Sets the operation name associated with this Call object
	void SetOperationName(const char *p_pchOpName);

	/// Set XML Mapping for user defined data types.
	void AddSerializerFactory(const std::string& p_sTypName, SerializerFactory * p_pSer);

	/// Sets the encoding style to the URL passed in.
	void SetEncodingStyle(const std::string& p_sNsURI);

	/**
     * Clears the list of parameters.
	 *
	 * @brief	Clears the list of parameters.
     */
	void ClearAllParams(){m_vArgs.clear();}


private:

	/// Clean up any memory associated with this call object
	void Cleanup();


private:
    std::vector<Param*> m_vArgs;			///< Vectore of arguments
    std::string         m_strOpName;		///< Operation/method name
    std::string         m_strMethodNsURI;	///< Method namespace URI
	std::string			m_ServerURL;		///< Remote address of the Web service
	std::string			m_EncStyle;			///< Encoding style

	SOAPMsgContext *m_pClientCtx;			///< SOAP Message handler

};

	

#endif // _AXIS_CALL_HPP


