/* -*- C++ -*- */

/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 2002 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "SOAP" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

/*
 *
 *
 * @author Lilantha Darshana (lilantha@erunway.com)
 *
 */
#include "Platform.hpp"
#include <iostream>
#include <cstdio>
#include "XMLDeSerializer.hpp"

#include <xercesc/util/PlatformUtils.hpp>
#include <xercesc/util/XMLUniDefs.hpp>
#include <xercesc/util/XMLException.hpp>
#include <xercesc/util/TranscodingException.hpp>
#include <xercesc/dom/DOM_DOMException.hpp>
#include <xercesc/parsers/DOMParser.hpp>
#include <xercesc/framework/MemBufInputSource.hpp>
#include "TypeMappingRegistry.hpp"


XMLDeSerializer::XMLDeSerializer(const std::string& p_Msg) : m_pTypeMapper(0)
{
	try
    {
        XMLPlatformUtils::Initialize();
    }
    catch (const XMLException& )
    {
        //TODO: need to handle exception 
    }

	parser = new DOMParser;
    parser->setValidationScheme(DOMParser::Val_Auto);
    parser->setDoNamespaces(true);
    parser->setDoSchema(false);
    parser->setValidationSchemaFullChecking(false);
    errReporter = new ParseErrorHandler();
    parser->setErrorHandler(errReporter);
    parser->setCreateEntityReferenceNodes(false);
    parser->setToCreateXMLDeclTypeNode(true);

	MemBufInputSource* memBufIS = new MemBufInputSource ((const XMLByte*) p_Msg.c_str(),
														  p_Msg.size(), "msg", false);
	bool errorsOccured = false;
	try
	{
		parser->parse(*memBufIS);
		int errorCount = parser->getErrorCount();
		if (errorCount > 0)
		{
			errorsOccured = true;
			m_pCurNode = NULL;
		}
	}
	catch (const XMLException& e)
	{
		//TODO: need to handle exception 
		std::cerr << "An error occured during parsing\n   Message: "
			<< DOMString(e.getMessage()).transcode() << std::endl;
		errorsOccured = true;
	}
	catch (const DOM_DOMException& e)
	{
		//TODO: need to handle exception 
		std::cerr << "A DOM error occured during parsing\n   DOMException code: "
			<< e.code << std::endl;
		errorsOccured = true;
	}
	catch (...)
	{
		//TODO: need to handle exception 
		std::cerr << "An error occured during parsing\n " << std::endl;
		errorsOccured = true;
	}


	if(!errorsOccured)
	{
		m_pCurNode = parser->getDocument();
	}
	else
		m_pCurNode = NULL;
}


XMLDeSerializer::~XMLDeSerializer()
{
	 // call the termination method
    XMLPlatformUtils::Terminate();
}


void XMLDeSerializer::DeSerialize(MessageElement& p_Elm)
{
	ax::QName qName = p_Elm.GetElmQName();
	std::string strQName = QNameToStr(qName);
	DOM_NodeList list = ((DOM_Document&)m_pCurNode).getElementsByTagName(strQName.c_str());

	unsigned int nNodeCnt = list.getLength();

	for(int i=0;i<nNodeCnt;i++)
	{
		DOM_Element data = (const DOM_Element&) list.item(i);
    
		DOM_Node child;
		for (child=data.getFirstChild(); child != 0; child= child.getNextSibling())
		{
			if(child.getNodeType() == DOM_Node::ELEMENT_NODE)
				p_Elm.AddMsgElement((DOM_Element&)child);
		}
	}
}


ax::QName XMLDeSerializer::GetQNameFromStr(std::string strQName)
{
	std::string::size_type pos = strQName.find(':');
	if (pos == std::string::npos) return ax::QName();
    
	std::string nsURI = m_nsStk.GetNamespaceURI(strQName.substr(0, pos));
    
    return ax::QName(nsURI, strQName.substr(++pos));
}

void XMLDeSerializer::RegisterPrefixForURI(const std::string& prefix, const std::string& uri)
{
	m_nsStk.Add(uri, prefix);
}

std::string XMLDeSerializer::QNameToStr(ax::QName& qName)
{
	std::string prefix = m_nsStk.GetPrefix(qName.GetNsURI());
	return ((!prefix.empty())?(prefix + ":"): std::string("")) + qName.GetLocalPart();
}



void ParseErrorHandler::error(const SAXParseException& e)
{
	//TODO: need to handle errors 
    fprintf(stderr, "\nError at file \"%s\", line %d, char %d:  %s\n",
        XMLString::transcode(e.getSystemId()), e.getLineNumber(),
        e.getColumnNumber(), XMLString::transcode(e.getMessage()));
    throw e;

};

void ParseErrorHandler::fatalError(const SAXParseException& e)
{
	//TODO: need to handle errors 
    fprintf(stderr, "\nFatal Error at file \"%s\", line %d, char %d:  %s\n",
        XMLString::transcode(e.getSystemId()), e.getLineNumber(),
        e.getColumnNumber(), XMLString::transcode(e.getMessage()));
    throw e;
};

void ParseErrorHandler::warning(const SAXParseException& e)
{
	//TODO: need to handle errors 
    fprintf(stderr, "\nWarning at file \"%s\", line %d, char %d:  %s\n",
        XMLString::transcode(e.getSystemId()), e.getLineNumber(),
        e.getColumnNumber(), XMLString::transcode(e.getMessage()));
    throw e;

};


