/* -*- C++ -*- */

/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 2002 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "SOAP" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

/*
 *
 *
 * @author Lilantha Darshana (lilantha@erunway.com)
 *
 */

#include "Platform.hpp"
#include "XMLSerializer.hpp"
#include "AttList.hpp"
#include "XercescConst.hpp"
#include "AxisConst.hpp"
#include "TypeMappingRegistry.hpp"
#include <cstdio>



std::ostream& operator<<(std::ostream& p_OutStr, const DOMString& p_dmNode);
std::ostream& operator<<(std::ostream& p_OutStr, DOM_Node& p_dmNode);
XMLFormatter& operator<< (XMLFormatter& p_xmlFmt, const DOMString& p_dmStr);

XMLFormatter* Xml_Formatter=NULL;
XMLCh* Encoding_Name=NULL;
static int m_nIndent=-1;




XMLSerializer::XMLSerializer(XmlWriter& str)
	: m_bStartDoc(true), m_nIndent(0), m_PrefixIndx(1), 
	  m_xmlBuffer(&str), m_pTypeMapper(0)
{
	try
    {
        XMLPlatformUtils::Initialize();
    }
    catch(const XMLException& ex)
    {
        char *pchMsg = XMLString::transcode(ex.getMessage());
		std::cerr << "Error during Xerces-c Initialization.\n"
				  << "  Exception message:"
                  << pchMsg;
        delete pchMsg;
    }
}

XMLSerializer::~XMLSerializer()
{
//	XMLPlatformUtils::Terminate();
}


void XMLSerializer::StartElem(ax::QName& p_qName, AttList& p_attList)
{
	// Start writing XML document
	if(m_bStartDoc)
	{
		m_Doc = DOM_Document::createDocument(); 
		m_Doc.appendChild(m_Doc.createXMLDecl ("1.0",  "UTF-8", ""));
		m_CurNode = m_Doc;
		Append(p_qName, p_attList);
		m_CurNode = m_Doc.getDocumentElement();
		m_bStartDoc = false;
		return;
	}

	m_CurNode = Append(p_qName, p_attList);
}


DOM_Element& XMLSerializer::Append(ax::QName& p_qName, AttList& p_attList)
{
	std::string qName = QNameToStr(p_qName);
	m_ChildElm = m_Doc.createElement(qName.c_str());

	for (int i = 0; i < p_attList.getLength(); i++) 
	{
		m_ChildElm.setAttribute(p_attList.getName(i).c_str(), 
								p_attList.getValue(i).c_str());
    }
	
	NsMap *CurMap = m_nsStk.Peek();
	if(CurMap)
	{
		NsMap::iterator mpIter;
		for (mpIter = CurMap->begin(); mpIter != CurMap->end(); ++mpIter) 
		{
			if(mpIter->first.empty()) continue;
			std::string  strQName = std::string("xmlns") + ":" + mpIter->second;
			m_ChildElm.setAttributeNS(ax::NS_URI_XMLNS.c_str(), strQName.c_str(),
										mpIter->first.c_str());
										
		}
		m_nsStk.Push();
	}
	
	m_CurNode.appendChild(m_ChildElm);
	return m_ChildElm;
}


void XMLSerializer::WriteStr(const std::string& str)
{
	m_CurTxtElm = m_Doc.createTextNode(str.c_str());
    m_ChildElm.appendChild(m_CurTxtElm);
}

void XMLSerializer::EndElem()
{
	m_nsStk.Pop();
	return;
}

void XMLSerializer::Serialize(ax::QName p_qName, AttList& p_attList, const char* value)
{
	m_ChildElm = Append(p_qName, p_attList);
	WriteStr(std::string(value));
}

void XMLSerializer::RegisterPrefixForURI(const std::string& prefix, const std::string& uri)
{
	m_nsStk.Add(uri, prefix);
}

std::string XMLSerializer::GetPrefixForURI(const std::string& uri)
{
	char buf[10];
	sprintf(buf, "%d", m_PrefixIndx++);
    return GetPrefixForURI(uri, "ns" + std::string(buf));
}

std::string XMLSerializer::GetPrefixForURI(const std::string& uri, std::string defaultPrefix)
{
    if (uri.empty()) 
        return std::string("");

	std::string prefix = m_nsStk.GetPrefix(uri);

    if (prefix.empty() && uri == ax::URI_SOAP_ENC)
	{
        prefix = ax::NSPREFIX_SOAP_ENC;
        RegisterPrefixForURI(prefix, uri);
    }

    if (prefix.empty()) 
	{
        prefix = defaultPrefix;
        RegisterPrefixForURI(prefix, uri);
    }

    return prefix; 
}

std::string XMLSerializer::QNameToStr(ax::QName& qName)
{
	std::string prefix = GetPrefixForURI(qName.GetNsURI());
	return ((!prefix.empty())?(prefix + ":"): std::string("")) + qName.GetLocalPart();
}


void XMLSerializer::flush()
{
	std::ostream p_OutStr(m_xmlBuffer->rdbuf());
    SOAPEnvelopFormatter* fmtTarget = new SOAPEnvelopFormatter(p_OutStr);

    DOMString xmlEncCharset("UTF-8");
    DOM_Node dmNode = m_Doc.getFirstChild();

	// Check whether this is a declaration Node 
    if (dmNode.getNodeType() == DOM_Node::XML_DECL_NODE)
    {
        DOMString strTmp = ((DOM_XMLDecl &)dmNode).getEncoding();
        if (strTmp != "")
        {
            xmlEncCharset = strTmp;
        }
    }

    unsigned int lent = xmlEncCharset.length();
    Encoding_Name	  = new XMLCh[lent + 1];

    XMLString::copyNString(Encoding_Name, xmlEncCharset.rawBuffer(), lent);
    Encoding_Name[lent] = 0;

    try
    {
        Xml_Formatter = new XMLFormatter(Encoding_Name, fmtTarget, XMLFormatter::NoEscapes, 
										 XMLFormatter::UnRep_CharRef);
        p_OutStr << m_Doc;
        *Xml_Formatter << chLF; // add linefeed in requested output encoding
    }
    catch (XMLException& e)
    {
		std::cerr << "An error occurred during creation of output transcoder. Msg is:"
				  << DOMString(e.getMessage()) << std::endl;
    }

    delete fmtTarget;
    delete Xml_Formatter;
}



std::ostream& operator<<(std::ostream& p_OutStr, DOM_Node& p_dmNode)
{
    // Get the name and value out for convenience
    DOMString   nodeName  = p_dmNode.getNodeName();
    DOMString   nodeValue = p_dmNode.getNodeValue();
    unsigned long lent = nodeValue.length();

    switch (p_dmNode.getNodeType())
    {
        case DOM_Node::TEXT_NODE:
		{
			Xml_Formatter->formatBuf(nodeValue.rawBuffer(), lent, XMLFormatter::CharEscapes);
			break;
		}

        case DOM_Node::PROCESSING_INSTRUCTION_NODE :
		{
			*Xml_Formatter << XMLFormatter::NoEscapes << gStartPI  << nodeName;

			if (lent > 0) 
				*Xml_Formatter << chSpace << nodeValue;

			*Xml_Formatter << XMLFormatter::NoEscapes << gEndPI;
			break;
		}

        case DOM_Node::DOCUMENT_NODE :
		{
			DOM_Node childNode = p_dmNode.getFirstChild();

			while( childNode != 0)
			{
				p_OutStr << childNode;
				*Xml_Formatter << chLF;	// add linefeed in requested output encoding
				p_OutStr << std::flush;
				childNode = childNode.getNextSibling();
			}
			break;
		}

        case DOM_Node::ELEMENT_NODE :
		{
			// The name has to be representable without any escapes
			m_nIndent++;	// make pretty
			std::string strTmp;

			if(m_nIndent > 0)
				strTmp = std::string(m_nIndent, ' ');
		
			// Output the element start tag.
			*Xml_Formatter  << XMLFormatter::NoEscapes
						 << strTmp.c_str() << chOpenAngle << nodeName;

			// Output any attributes on this element
			DOM_NamedNodeMap mapAttribs = p_dmNode.getAttributes();
			int nAttCnt = mapAttribs.getLength();

			for (int i = 0; i < nAttCnt; i++)
			{
				DOM_Node  dnAttr = mapAttribs.item(i);

				//  Again the name has to be completely representable. But the
				//  dnAttr can have refs and requires the dnAttr style
				//  escaping.
				*Xml_Formatter	 << XMLFormatter::NoEscapes
								 << chSpace << dnAttr.getNodeName()
								 << chEqual << chDoubleQuote
								 << XMLFormatter::AttrEscapes
								 << dnAttr.getNodeValue()
								 << XMLFormatter::NoEscapes
								 << chDoubleQuote;
			}

			//  Test for the presence of children, which includes both
			//  text content and nested elements.
			DOM_Node childNode = p_dmNode.getFirstChild();

			if (childNode != 0)
			{
				// There are children. Close start-tag, and output children.
				// No escapes are legal here
				bool bTxtChild=false;

				*Xml_Formatter << XMLFormatter::NoEscapes << chCloseAngle;

				//make pretty
				if(childNode.getNodeType() != DOM_Node::TEXT_NODE)
					*Xml_Formatter << "\n";

				while( childNode != 0)
				{
					if(childNode.getNodeType() == DOM_Node::TEXT_NODE)
						bTxtChild = true;

					p_OutStr << childNode;
					childNode = childNode.getNextSibling();
				}

				// Done with children.  Output the end tag.
				if(!bTxtChild && (m_nIndent > 0))
					strTmp = std::string(m_nIndent, ' ');		//make pretty
				else
					strTmp = "";
			
				*Xml_Formatter << strTmp.c_str() << XMLFormatter::NoEscapes << gEndElement
							<< nodeName << chCloseAngle << "\n";
				m_nIndent--;
			}
			else
			{
				//  There were no children. Output the short form close of
				//  the element start tag, making it an empty-element tag.
				*Xml_Formatter << XMLFormatter::NoEscapes << chForwardSlash << chCloseAngle << "\n";
			}
			break;
		}


        case DOM_Node::ENTITY_REFERENCE_NODE:
		{
			DOM_Node childNode;

			   *Xml_Formatter << XMLFormatter::NoEscapes << chAmpersand
					<< nodeName << chSemiColon;
			break;
		}


        case DOM_Node::CDATA_SECTION_NODE:
		{
			*Xml_Formatter << XMLFormatter::NoEscapes << gStartCDATA
						<< nodeValue << gEndCDATA;
			break;
		}

        case DOM_Node::XML_DECL_NODE:
		{
			DOMString  str;

			*Xml_Formatter << gXMLDecl1 << ((DOM_XMLDecl &)p_dmNode).getVersion();

			*Xml_Formatter << gXMLDecl2 << Encoding_Name;

			str = ((DOM_XMLDecl &)p_dmNode).getStandalone();
			if (str != 0)
				*Xml_Formatter << gXMLDecl3 << str;

			*Xml_Formatter << gXMLDecl4;

			break;
		}

        default:
            std::cerr << "Unrecognized node type for SOAP envelop = "
				<< (long)p_dmNode.getNodeType() << std::endl;
    }
    return p_OutStr;
}

std::ostream& operator<< (std::ostream& p_OutStr, const DOMString& p_dmStr)
{
    char *p = p_dmStr.transcode();
    p_OutStr << p;
    delete [] p;
    return p_OutStr;
}


XMLFormatter& operator<< (XMLFormatter& p_xmlFmt, const DOMString& p_dmStr)
{
    unsigned int lent = p_dmStr.length();

	if (lent <= 0)
		return p_xmlFmt;

    XMLCh*  buf = new XMLCh[lent + 1];
    XMLString::copyNString(buf, p_dmStr.rawBuffer(), lent);
    buf[lent] = 0;
    p_xmlFmt << buf;
    delete [] buf;
    return p_xmlFmt;
}







        
