/* -*- C++ -*- */

/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 2002 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "SOAP" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

/*
 *
 *
 * @author Lilantha Darshana (lilantha@erunway.com)
 *
 */


#if !defined(_AXIS_XML_SERIALIZER_HPP)
#define _AXIS_XML_SERIALIZER_HPP


#include <xercesc/util/PlatformUtils.hpp>
#include <xercesc/util/XMLString.hpp>
#include <xercesc/util/XMLUniDefs.hpp>
#include <xercesc/dom/DOM.hpp>
#include <xercesc/framework/XMLFormatter.hpp>
#include <xercesc/util/TranscodingException.hpp>
#include <stack>
#include <iostream>
#include "XmlWriter.hpp"
#include "QName.hpp"
#include "NsStack.hpp"
#include "AttList.hpp"

class TypeMappingRegistry;

/**
 *	Manage a serialization, including keeping track of namespace mappings 
 *	and element stacks. 
 *	
 * @brief  Serializes XML SOAP Envelop fragments to form a SOAP Message. 
 *
 */

class XMLSerializer
{
	public:
		XMLSerializer(XmlWriter& str);
		~XMLSerializer();

		/// Register a Prefix for a namespace URI
		void RegisterPrefixForURI(const std::string& prefix, const std::string& uri);

		/// Get a prefix for a given namespace URI
		std::string GetPrefixForURI(const std::string& uri);

		/// Get a prefix for the given namespace URI 
		std::string GetPrefixForURI(const std::string& uri, std::string defaultPrefix);

		/// Convert Qualified Name to a string as: prefix:local-name
		std::string QNameToStr(ax::QName& qName);

		/// Sets a reference to the TypeMapping Registry 
		void SetTypeMappingRegistry(TypeMappingRegistry* p_TypMapper){m_pTypeMapper = p_TypMapper;}

		/// Obtain a reference to the TypeMappingRegistry we're using
		TypeMappingRegistry * GetTypeMapper(){return m_pTypeMapper;}

		/// Start building DOM sub-tree rooted at element having tag name with the given QName
		void StartElem(ax::QName& p_qName, AttList& p_attList);

		/// End this branch of the DOM tree.
		void EndElem();

		/// Write TEXT- node
		void WriteStr(const std::string& str);

		/// Append a subtree to the current node.
		DOM_Element& Append(ax::QName& p_qName, AttList& p_attList);

		/// Serialize a SOAP body element by appending a branch to the current node
		void Serialize(ax::QName p_qName, AttList& p_attList, const char* value);

		/// Write the content of the DOM document into a stream.
		void flush();

	private:

		friend std::ostream& operator<<(std::ostream& target, const DOMString& toWrite);
		friend std::ostream& operator<<(std::ostream& target, DOM_Node& toWrite);
		friend XMLFormatter& operator<<(XMLFormatter& strm, const DOMString& s);


	private:
		NsStack	m_nsStk;		///< Namespace stack for namespace mapping for the current context

		bool m_bStartDoc;		///< Starting to build a DOM document from SOAP Message components.
		int  m_nIndent;			///< To make pertty.
		int  m_PrefixIndx;		///< To make pretty.

		XmlWriter *m_xmlBuffer;	///< output stream to dump serilized DOM document.

		TypeMappingRegistry *m_pTypeMapper;

		// Use DOM to build the SOAP Envelop document
		DOM_Document m_Doc;
		DOM_Node     m_CurNode;
		DOM_Element	 m_CurElm;
		DOM_Element	 m_ChildElm;
		DOM_Text	 m_CurTxtElm;


};

/**
 * 
 * The implementation of the XMLFormatTarget to dump generated SOAP Envelop as a 
 * DOM tree, into a output stream. For more info refer Xerces-C++ DOM parser 
 * documentation.
 *	
 * @brief  Implements XMLFormatTarget for dumping generated DOM tree into a output stream.
 *
 */

class SOAPEnvelopFormatter : public XMLFormatTarget
{
public:
    SOAPEnvelopFormatter(std::ostream& p_xmlBuffer) : m_xmlBuf(&p_xmlBuffer) {}
    ~SOAPEnvelopFormatter() {};

    void writeChars(const   XMLByte* const  p_chWhat,
                    const   unsigned int    p_Cnt,
                            XMLFormatter * const p_xmlFormatter)
    {
		m_xmlBuf->write((char *) p_chWhat, (int) p_Cnt);
    };

private:
    SOAPEnvelopFormatter(const SOAPEnvelopFormatter& p_Other);
    void operator=(const SOAPEnvelopFormatter& p_rvalue);

	std::ostream *m_xmlBuf;
};

#endif //_AXIS_XML_SERIALIZER_HPP

